/*
 * changepw - Allow a user to change his password on an LDAP server.  
 *            See also changepw.html.
 * 
 * This program is for demonstration purposes only
 * 
 * Based on code Copyright 1999 Netscape Communications Corporation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdarg.h>

#include <ldap.h>

#define LDAP_HOST	"localhost"
#define LDAP_PORT	389

#define BASE "o=mcom.com"

/* Form handling functions. */
static char **input;

void report_error(char const * fmt, ...) ;
void form_unescape(char * str);
void post_begin(FILE * in);
char **string_to_vec(char * in);
char *get_cgi_var(char const * varname);
char *escape_for_shell(char * cmd);

char * find_user(char * filter);
int change_password(char * user, char * old_pw, char * new_pw);

int main(int argc, char *argv[])
{
    char *m = getenv("REQUEST_METHOD");
    
    fprintf(stdout, "Content-type: text/html\n\n");

    if (!m)
	report_error("You should only use the POST method with this program.");

    if (strcmp(m, "GET")==0)  {
        report_error("You should only use the POST method with this program.");
    }  else  {
        char *user;
        char *old_pw;
        char *new_pw1;
        char *new_pw2;

        post_begin(stdin);

        user = get_cgi_var("user");
        if (!user)  {
            user = getenv("REMOTE_USER");
            if (!user)
                report_error("Could not determine your username.  Please enter it on the form provided.");
        }

        old_pw = get_cgi_var("old_pw");
        if (!old_pw)
            report_error("You need to enter your old password for verification.");
        new_pw1 = get_cgi_var("new_pw1");
        if (!new_pw1)
            report_error("Be sure to enter both new passwords.");

        new_pw2 = get_cgi_var("new_pw2");
        if (!new_pw2)
            report_error("Be sure to enter both new passwords.");

        if (strcmp(new_pw1, new_pw2))  
            report_error("The new passwords you gave did not match.");

	if (change_password(user, old_pw, new_pw1)!=0)
            report_error("An unexpected error occured (sorry, no additional details).");

        fprintf(stdout, 
              "<hr width=50%%><h1 align=center>Success!</h1><hr width=50%%>\n");
        fprintf(stdout,"<p align=center>Your password has been changed.</p>\n");
    } 
    return 0;
}

char * find_user(char * filter)
{
    LDAP    	*ld;
    LDAPMessage	*result, *e;
    char		*dn=NULL;
    int		numentries=0;
    int		rc;

    /* get a handle to an LDAP connection */
    if ((ld = ldap_init( LDAP_HOST, LDAP_PORT)) == NULL) {
  	report_error("ldap_init: Unable to initiate connection to LDAP server %s on port %d", LDAP_HOST, LDAP_PORT);
    }

    /* authenticate to the directory as nobody */
    rc = ldap_simple_bind_s(ld, NULL, NULL);
    if (rc != LDAP_SUCCESS) {
    	report_error("ldap_simple_bind_s: %s", ldap_err2string(rc));
    }

    /* search for all entries with our filter */
    rc = ldap_search_s(ld, BASE, LDAP_SCOPE_SUBTREE,
    	filter, NULL, 0, &result);
    if (rc != LDAP_SUCCESS)
    	report_error("ldap_search_s: %s", ldap_err2string(rc));

    /* for each entry grab the DN, if there is more than 1 we'll error out */
    for (e = ldap_first_entry( ld, result); e != NULL;
        e = ldap_next_entry(ld, e)) {
    	if ((dn = ldap_get_dn( ld, e)) != NULL) {
    		/* do nothing */ ;
    	}
    	numentries++;
    }

    ldap_msgfree(result);

    if (numentries !=1)
    	report_error("A single entry was not found when searching.");

    ldap_unbind(ld);
    return(dn);
}

int change_password(char * user, char * old_pw, char * new_pw)
{
    char	searchdn[64];
    char	*binddn;
    LDAP	*ld;
    LDAPMod	mod;
    LDAPMod	*mods[ 2 ];
    char	*vals[ 2 ];
    int		rc;

    /* set up our search filter */
    snprintf(searchdn, 64, "uid=%s", user);
    binddn = find_user(searchdn);

    /* get a handle to an LDAP connection */
    if ((ld = ldap_init( LDAP_HOST, LDAP_PORT)) == NULL) {
	report_error("ldap_init: Unable to initiate connection to LDAP server %s on port %d", LDAP_HOST, LDAP_PORT);
    }

    /* authenticate to server as the user we found */
    rc = ldap_simple_bind_s(ld, binddn, old_pw);
    if (rc != LDAP_SUCCESS)
	report_error("ldap_simple_bind_s: %s", ldap_err2string(rc));

    /* construct the list of modifications to make */
    mod.mod_op = LDAP_MOD_REPLACE;
    mod.mod_type = "userpassword";
    vals[0] = new_pw;
    vals[1] = NULL;
    mod.mod_values = vals;

    mods[ 0 ] = &mod;
    mods[ 1 ] = NULL;

    /* make the change */
    rc = ldap_modify_s(ld, binddn, mods);
    if (rc != LDAP_SUCCESS)
	report_error("ldap_modify_s: %s", ldap_err2string(rc));

    ldap_unbind(ld);
    return(0);
}


/* Tell the user about an unfortunate occurrence. */
void report_error(char const * fmt, ...) 
{

    va_list args;

    fprintf(stdout, "<html><body><hr size=3><h1 align=center>Error</h1><hr size=3><b>\n");
    va_start(args,fmt);
    vfprintf(stdout,fmt,args);
    va_end(args);

    fprintf(stdout, "</b></body></html>\n");
    exit(1);
}

/* Unescape the %xx variables as they're sent in. */
void form_unescape(char *str) 
{
    register int x = 0, y = 0;
    int l = strlen(str);
    char digit;

    while(x < l)  {
        if ((str[x] == '%') && (x < (l - 2)))  {
            ++x;
            digit = (str[x] >= 'A' ? 
                         ((str[x] & 0xdf) - 'A')+10 : (str[x] - '0'));
            digit *= 16;

            ++x;
            digit += (str[x] >= 'A' ? 
                         ((str[x] & 0xdf) - 'A')+10 : (str[x] - '0'));

            str[y] = digit;
        } 
        else if (str[x] == '+')  {
            str[y] = ' ';
        } else {
            str[y] = str[x];
        }
        x++;
        y++;
    }
    str[y] = '\0';
}

/* Read in the variables from stdin, unescape them, and then put them in 
 * the static vector. */
void post_begin(FILE *in) 
{
    char *vars = NULL, *tmp = NULL;
    int cl;

    if (!(tmp = getenv("CONTENT_LENGTH")))
        report_error("Your browser sent no content length with a POST command."
                     "  Please be sure to use a fully compliant browser.");
        
    cl = atoi(tmp);

    vars = (char *)malloc(cl+1);

    if (!(fread(vars, 1, cl, in)))
        report_error("The POST variables could not be read from stdin.");

    vars[cl] = '\0';

    input = string_to_vec(vars);
}

/* Convert the input from stdin to a usable variable vector. */
char **string_to_vec(char *in)
{
    char **ans;
    int vars = 0;
    register int x = 0;
    char *tmp;
    
    while(in[x])
        if (in[x++]=='=')
            vars++;
    
    ans = (char **) malloc((sizeof(char *)) * (vars+1));
  
    x=0;
    tmp = strtok(in, "&");
    ans[x]=strdup(tmp);
    form_unescape(ans[x++]);

    while((tmp = strtok(NULL, "&")))  {
        ans[x] = strdup(tmp);
        form_unescape(ans[x++]);
    }
    return(ans);
}

/* Return the value of a POSTed variable, or NULL if none was sent. */
char *get_cgi_var(char const * varname)
{
    register int x = 0;
    int len = strlen(varname);
    char *ans = NULL;
   
    while(input[x])  {
    /*  We want to get rid of the =, so len, len+1 */
        if ((!strncmp(input[x], varname, len)) && (*(input[x]+len) == '='))  {
            ans = strdup(input[x] + len + 1);
            if (!strcmp(ans, ""))
                ans = NULL;
            break;
        }  else
            x++;
    }
    return ans;
}

/* Escape any shell-specific characters with a "\" */
char *escape_for_shell(char *cmd) {
    register int x,y,l;
    char *ans=(char *)malloc(2*strlen(cmd)+2);
 
    strcpy(ans, cmd);
    l=strlen(ans);
    for(x=0;ans[x];x++) {
        if (strchr(" &;`'\"|*!?~<>^()[]{}$\\",ans[x])){
            for(y=l+1;y>x;y--)
                ans[y] = ans[y-1];
            l++; /* length has been increased */
            ans[x] = '\\';
            x++; /* skip the character */
        }
    }
    return ans;
}
